#!/bin/sh

# Copyright 2017 - 2018 Paul Hänsch
#
# This is CGIlite.
# A collection of posix shell functions for writing CGI scripts.
# 
# CGIlite is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# CGIlite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with CGIlite.  If not, see <http://www.gnu.org/licenses/>. 

# ksh and zsh workaround
# set -o posix # ksh, not portable
setopt -o OCTAL_ZEROES 2>&-

BR="$(printf '\n')"
CR="$(printf '\r')"
cgilite_timeout=2

HEADER(){
  # Read value of header line. Use this instead of
  # referencing HTTP_* environment variables.
  if [ -n "${cgilite_headers+x}" ]; then
    printf %s "$cgilite_headers" \
    | sed -rn 's;^'"${1}"': ([^\r]+)\r?$;\1;i; tX; d; :X;p;q;'
  else
    eval "printf %s \"\$HTTP_$(printf %s "${1}" |tr a-z A-Z |tr -c A-Z _)\""
  fi
}

HEX_DECODE(){
  printf "$(printf %s "$1" \
  | sed -r '
    s;\\;\\\\;g; :x; s;%([^0-9A-F]);\\045\1;g; tx;
    # Hexadecimal { %00 - %FF } will be transformed to octal { \000 - \377 } for posix printf
    s;%[0123].;&\\0;g; s;%[4567].;&\\1;g; s;%[89AB].;&\\2;g; s;%[CDEF].;&\\3;g;
    s;%[048C][0-7]\\.;&0;g; s;%[048C][89A-F]\\.;&1;g; s;%[159D][0-7]\\.;&2;g; s;%[159D][89A-F]\\.;&3;g;
    s;%[26AE][0-7]\\.;&4;g; s;%[26AE][89A-F]\\.;&5;g; s;%[37BF][0-7]\\.;&6;g; s;%[37BF][89A-F]\\.;&7;g;
    s;%.[08](\\..);\10;g; s;%.[19](\\..);\11;g; s;%.[2A](\\..);\12;g; s;%.[3B](\\..);\13;g;
    s;%.[4C](\\..);\14;g; s;%.[5D](\\..);\15;g; s;%.[6E](\\..);\16;g; s;%.[7F](\\..);\17;g;
  ')"
}

if [ -z "$REQUEST_METHOD" -a -z "$SERVER_PROTOCOL" ]; then
  # no webserver variables means we are running via inetd / ncat
  # so use builtin web server

  # Use env from inetd as webserver variables
  REMOTE_ADDR="${TCPREMOTEIP}"
  SERVER_NAME="${TCPLOCALIP}"
  SERVER_PORT="${TCPLOCALPORT}"

  # Wait 2 seconds for request or kill connection through watchdog.
  # Once Request is received the watchdog will be suspended (killed).
  # At the end of the loop the watchdog will be restarted to enable
  # timeout for the subsequent request.

  (sleep $cgilite_timeout && kill $$) & cgilite_watchdog=$!
  while read REQUEST_METHOD REQUEST_URI SERVER_PROTOCOL; do
    kill $cgilite_watchdog
    PATH_INFO="$(HEX_DECODE "${REQUEST_URI%\?*}")"
    QUERY_STRING="${REQUEST_URI#*\?}"
    cgilite_headers="$(sed -u '/^\r\?$/q')"

    HTTP_CONTENT_LENGTH="$(HEADER Content-Length |grep -xE '[0-9]+')"

    export REMOTE_ADDR SERVER_NAME SERVER_PORT REQUEST_METHOD REQUEST_URI SERVER_PROTOCOL \
           PATH_INFO QUERY_STRING HTTP_CONTENT_LENGTH

    # Try to serve multiple requests, provided that script serves a
    # Content-Length header.
    # Without Content-Length header, connection will terminate after
    # script.

    . "$0" |sed '1{s;^Status: ;HTTP/1.1 ;; t; s;^;HTTP/1.1 200 OK\r\n;;}' \
    | while read -r l; do case $l in
      Content-Length:*) printf '%s\n' "$l"; cat;;
      $CR) printf 'Connection: close\r\n\r\n'; cat; exit 1;;
      *) printf '%s\n' "$l";;
    esac; done || exit 0;
    (sleep $cgilite_timeout && kill $$) & cgilite_watchdog=$!
  done
fi

if [ "$REQUEST_METHOD" = POST -a "${HTTP_CONTENT_LENGTH:=${CONTENT_LENGTH:=0}}" -gt 0 ]; then
  cgilite_post="$(head -c "$HTTP_CONTENT_LENGTH")"
fi

[ -n "${DEBUG+x}" ] && env

cgilite_count(){
  case $1 in
    GET)  printf %s "&${QUERY_STRING}";;
    POST) printf %s "&${cgilite_post}";;
    REF)  printf %s "&${HTTP_REFERER#*\?}";;
  esac \
  | grep -Eo '&'"$2"'=[^&]*' \
  | wc -l
}

cgilite_value(){
  HEX_DECODE "$(
    case $1 in
      GET)  printf %s "&${QUERY_STRING}";;
      POST) printf %s "&${cgilite_post}";;
      REF)  printf %s "&${HTTP_REFERER#*\?}";;
    esac \
    | grep -Eo '&'"$2"'=[^&]*' \
    | sed -rn "${3:-1}"'{s;^[^=]+=;;; s;\+; ;g; p;}'
  )"
}

GET(){ cgilite_value GET $@; }
GET_COUNT(){ cgilite_count GET $1; }

POST(){ cgilite_value POST $@; }
POST_COUNT(){ cgilite_count POST $1; }

REF(){ cgilite_value REF $@; }
REF_COUNT(){ cgilite_count REF $1; }

COOKIE(){
  HEX_DECODE "$(
    HEADER Cookie \
    | grep -oE '(^|; ?)'"$1"'=[^;]*' \
    | sed -rn "${2:-1}"'{s;^[^=]+=;;; s;\+; ;g; p;}'
  )"
}

HTML(){
  # HTML Entity Coding
  # Prints UTF-8 string as decimal Unicode Code Points
  # Useful for escaping user input for use in HTML text and attributes
  printf %s "$*" \
  | hexdump -ve '/1 "%03o\n"' \
  | while read n; do
    case $n in
      # bitbanging octal UTF-8 chains into singular 7 digit octal numbers
      [01]??) printf '0000%s' $n;; # 7 bit ASCII character, nothing to do
      2??)    printf '%s' ${n#2};; # tail fragment, append 6 bit
      3[0123]?) printf '000%s' ${n#3};; # 2 octet (11 bit) chain start
      34?) printf '00%s' ${n#34};; # 3 octet (16 bit) chain start
      35?) printf '01%s' ${n#35};; # 3 octet chain start, high
      36?) printf '%s' ${n#36};;   # 4 octet (21 bit) chain start
    esac
  done \
  | sed -r 's;.{7};&\n;g;' \
  | while read n; do
    printf '&#%d;' $((0$n))
  done
}

URL(){
  # Code every character in URL escape hex format
  # except alphanumeric ascii

  printf %s "$*" \
  | hexdump -v -e '/1 ",%02X"' \
  | tr , %
}

SET_COOKIE(){
  case "$1" in
    ''|0|session) expire='';;
    [+-][0-9]*)   expire="$(date -R -d @$(($(date +%s) + $1)))";;
    *)		  expire="$(date -R -d "$1")";;
  esac
  cookie="$2"
  
  printf 'Set-Cookie: %s' "$cookie"
  [ -n "$expire" ] && printf '; Expires=%s' "$expire" 
  [ $# -ge 3 ] && shift 2 && printf '; %s' "$@"
  printf '\r\n'
}

REDIRECT(){
  printf 'Status: 303 See Other\r\nLocation: %s\r\n\r\n' "$*"
  exit 0
}

